import graphene
from dagster._core.definitions.freshness import InternalFreshnessPolicy, TimeWindowFreshnessPolicy


class GrapheneTimeWindowFreshnessPolicy(graphene.ObjectType):
    class Meta:
        name = "TimeWindowFreshnessPolicy"

    failWindowSeconds = graphene.NonNull(graphene.Int)
    warnWindowSeconds = graphene.Int()


class GrapheneInternalFreshnessPolicy(graphene.Union):
    class Meta:
        name = "InternalFreshnessPolicy"
        types = (GrapheneTimeWindowFreshnessPolicy,)

    @classmethod
    def from_policy(cls, policy: InternalFreshnessPolicy):
        if isinstance(policy, TimeWindowFreshnessPolicy):
            return GrapheneTimeWindowFreshnessPolicy(
                failWindowSeconds=policy.fail_window.to_timedelta().total_seconds(),
                warnWindowSeconds=policy.warn_window.to_timedelta().total_seconds()
                if policy.warn_window
                else None,
            )
        raise Exception("Unknown freshness policy type")



def types():
    from dagster_graphql.schema.pipelines.config import (
        GrapheneEvaluationErrorReason,
        GrapheneEvaluationStack,
        GrapheneEvaluationStackEntry,
        GrapheneEvaluationStackListItemEntry,
        GrapheneEvaluationStackMapKeyEntry,
        GrapheneEvaluationStackMapValueEntry,
        GrapheneEvaluationStackPathEntry,
        GrapheneFieldNotDefinedConfigError,
        GrapheneFieldsNotDefinedConfigError,
        GrapheneMissingFieldConfigError,
        GrapheneMissingFieldsConfigError,
        GraphenePipelineConfigValidationError,
        GraphenePipelineConfigValidationInvalid,
        GraphenePipelineConfigValidationValid,
        GrapheneRunConfigValidationInvalid,
        GrapheneRuntimeMismatchConfigError,
        GrapheneSelectorTypeConfigError,
    )
    from dagster_graphql.schema.pipelines.config_result import (
        GraphenePipelineConfigValidationResult,
    )
    from dagster_graphql.schema.pipelines.logger import GrapheneLogger
    from dagster_graphql.schema.pipelines.mode import GrapheneMode
    from dagster_graphql.schema.pipelines.pipeline import (
        GrapheneAsset,
        GrapheneIPipelineSnapshot,
        GraphenePipeline,
        GraphenePipelinePreset,
        GraphenePipelineRun,
        GrapheneRun,
        GrapheneRunOrError,
    )
    from dagster_graphql.schema.pipelines.pipeline_ref import (
        GraphenePipelineReference,
        GrapheneUnknownPipeline,
    )
    from dagster_graphql.schema.pipelines.pipeline_run_stats import (
        GraphenePipelineRunStatsSnapshot,
        GrapheneRunStatsSnapshot,
        GrapheneRunStatsSnapshotOrError,
    )
    from dagster_graphql.schema.pipelines.resource import GrapheneResource
    from dagster_graphql.schema.pipelines.snapshot import (
        GraphenePipelineSnapshot,
        GraphenePipelineSnapshotOrError,
    )
    from dagster_graphql.schema.pipelines.status import GrapheneRunStatus
    from dagster_graphql.schema.pipelines.subscription import (
        GraphenePipelineRunLogsSubscriptionFailure,
        GraphenePipelineRunLogsSubscriptionPayload,
        GraphenePipelineRunLogsSubscriptionSuccess,
    )

    return [
        GrapheneAsset,
        GrapheneEvaluationErrorReason,
        GrapheneEvaluationStack,
        GrapheneEvaluationStackEntry,
        GrapheneEvaluationStackListItemEntry,
        GrapheneEvaluationStackPathEntry,
        GrapheneEvaluationStackMapKeyEntry,
        GrapheneEvaluationStackMapValueEntry,
        GrapheneFieldNotDefinedConfigError,
        GrapheneFieldsNotDefinedConfigError,
        GrapheneIPipelineSnapshot,
        GrapheneLogger,
        GrapheneMissingFieldConfigError,
        GrapheneMissingFieldsConfigError,
        GrapheneMode,
        GraphenePipeline,
        GraphenePipelineConfigValidationError,
        GraphenePipelineConfigValidationInvalid,
        GrapheneRunConfigValidationInvalid,
        GraphenePipelineConfigValidationResult,
        GraphenePipelineConfigValidationValid,
        GraphenePipelinePreset,
        GraphenePipelineReference,
        GraphenePipelineRun,
        GraphenePipelineRunLogsSubscriptionFailure,
        GraphenePipelineRunLogsSubscriptionPayload,
        GraphenePipelineRunLogsSubscriptionSuccess,
        GrapheneRunOrError,
        GraphenePipelineRunStatsSnapshot,
        GrapheneRunStatsSnapshotOrError,
        GrapheneRunStatsSnapshot,
        GrapheneRunStatus,
        GraphenePipelineSnapshot,
        GraphenePipelineSnapshotOrError,
        GrapheneResource,
        GrapheneRuntimeMismatchConfigError,
        GrapheneRun,
        GrapheneSelectorTypeConfigError,
        GrapheneUnknownPipeline,
    ]



import graphene

from dagster_graphql.schema.errors import GrapheneError
from dagster_graphql.schema.util import non_null_list


class GraphenePartitionKeyConnection(graphene.ObjectType):
    results = non_null_list(graphene.String)
    cursor = graphene.NonNull(graphene.String)
    hasMore = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "PartitionKeyConnection"


class GraphenePartitionKeys(graphene.ObjectType):
    partitionKeys = non_null_list(graphene.String)

    class Meta:
        name = "PartitionKeys"


class GraphenePartitionSubsetDeserializationError(graphene.ObjectType):
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneError,)
        name = "PartitionSubsetDeserializationError"


class GraphenePartitionKeysOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitionKeys, GraphenePartitionSubsetDeserializationError)
        name = "PartitionKeysOrError"



import datetime
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, Any, Optional, Union  # noqa: UP035

from dagster import (
    AssetKey,
    _check as check,
)
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.errors import DagsterInvariantViolationError, DagsterRunNotFoundError
from dagster._core.execution.backfill import BulkActionsFilter, BulkActionStatus
from dagster._core.instance import DagsterInstance
from dagster._core.storage.dagster_run import DagsterRunStatus, RunRecord, RunsFilter
from dagster._core.storage.event_log.base import AssetRecord
from dagster._core.storage.tags import BACKFILL_ID_TAG, TagType, get_tag_type
from dagster._record import copy, record
from dagster._time import datetime_from_timestamp
from dagster._utils.warnings import disable_dagster_warnings

from dagster_graphql.implementation.external import ensure_valid_config, get_remote_job_or_raise

if TYPE_CHECKING:
    from dagster_graphql.schema.asset_graph import GrapheneAssetLatestInfo
    from dagster_graphql.schema.errors import GrapheneRunNotFoundError
    from dagster_graphql.schema.execution import GrapheneExecutionPlan
    from dagster_graphql.schema.logs.events import GrapheneRunStepStats
    from dagster_graphql.schema.pipelines.config import GraphenePipelineConfigValidationValid
    from dagster_graphql.schema.pipelines.pipeline import GrapheneEventConnection, GrapheneRun
    from dagster_graphql.schema.pipelines.pipeline_run_stats import GrapheneRunStatsSnapshot
    from dagster_graphql.schema.runs import GrapheneRunGroup, GrapheneRunTagKeys, GrapheneRunTags
    from dagster_graphql.schema.runs_feed import GrapheneRunsFeedConnection, GrapheneRunsFeedView
    from dagster_graphql.schema.util import ResolveInfo


_DELIMITER = "::"


async def gen_run_by_id(
    graphene_info: "ResolveInfo", run_id: str
) -> Union["GrapheneRun", "GrapheneRunNotFoundError"]:
    from dagster_graphql.schema.errors import GrapheneRunNotFoundError
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

    record = await RunRecord.gen(graphene_info.context, run_id)
    if not record:
        return GrapheneRunNotFoundError(run_id)
    else:
        return GrapheneRun(record)


def get_run_tag_keys(graphene_info: "ResolveInfo") -> "GrapheneRunTagKeys":
    from dagster_graphql.schema.runs import GrapheneRunTagKeys

    return GrapheneRunTagKeys(
        keys=[
            tag_key
            for tag_key in graphene_info.context.instance.get_run_tag_keys()
            if get_tag_type(tag_key) != TagType.HIDDEN
        ]
    )


def get_run_tags(
    graphene_info: "ResolveInfo",
    tag_keys: list[str],
    value_prefix: Optional[str] = None,
    limit: Optional[int] = None,
) -> "GrapheneRunTags":
    from dagster_graphql.schema.runs import GrapheneRunTags
    from dagster_graphql.schema.tags import GraphenePipelineTagAndValues

    instance = graphene_info.context.instance
    return GrapheneRunTags(
        tags=[
            GraphenePipelineTagAndValues(key=key, values=values)
            for key, values in instance.get_run_tags(
                tag_keys=tag_keys, value_prefix=value_prefix, limit=limit
            )
            if get_tag_type(key) != TagType.HIDDEN
        ]
    )


def get_run_group(graphene_info: "ResolveInfo", run_id: str) -> "GrapheneRunGroup":
    from dagster_graphql.schema.errors import GrapheneRunGroupNotFoundError
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
    from dagster_graphql.schema.runs import GrapheneRunGroup

    instance = graphene_info.context.instance
    try:
        result = instance.get_run_group(run_id)
        if result is None:
            return GrapheneRunGroupNotFoundError(run_id)
    except DagsterRunNotFoundError:
        return GrapheneRunGroupNotFoundError(run_id)
    root_run_id, run_group = result
    run_group_run_ids = [run.run_id for run in run_group]
    records_by_id = {
        record.dagster_run.run_id: record
        for record in instance.get_run_records(RunsFilter(run_ids=run_group_run_ids))
    }
    return GrapheneRunGroup(
        root_run_id=root_run_id,
        runs=[GrapheneRun(records_by_id[run_id]) for run_id in run_group_run_ids],
    )


def get_runs(
    graphene_info: "ResolveInfo",
    filters: Optional[RunsFilter],
    cursor: Optional[str] = None,
    limit: Optional[int] = None,
) -> Sequence["GrapheneRun"]:
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

    check.opt_inst_param(filters, "filters", RunsFilter)
    check.opt_str_param(cursor, "cursor")
    check.opt_int_param(limit, "limit")

    instance = graphene_info.context.instance

    return [
        GrapheneRun(record)
        for record in instance.get_run_records(filters=filters, cursor=cursor, limit=limit)
    ]


def get_run_ids(
    graphene_info: "ResolveInfo",
    filters: Optional[RunsFilter],
    cursor: Optional[str] = None,
    limit: Optional[int] = None,
) -> Sequence[str]:
    check.opt_inst_param(filters, "filters", RunsFilter)
    check.opt_str_param(cursor, "cursor")
    check.opt_int_param(limit, "limit")

    instance = graphene_info.context.instance

    return instance.get_run_ids(filters=filters, cursor=cursor, limit=limit)


PENDING_STATUSES = [
    DagsterRunStatus.STARTING,
    DagsterRunStatus.MANAGED,
    DagsterRunStatus.NOT_STARTED,
    DagsterRunStatus.QUEUED,
    DagsterRunStatus.STARTED,
    DagsterRunStatus.CANCELING,
]
IN_PROGRESS_STATUSES = [
    DagsterRunStatus.STARTED,
    DagsterRunStatus.CANCELING,
]


def _get_latest_planned_run_id(instance: DagsterInstance, asset_record: AssetRecord):
    if instance.event_log_storage.asset_records_have_last_planned_and_failed_materializations:
        return asset_record.asset_entry.last_planned_materialization_run_id
    else:
        planned_info = instance.get_latest_planned_materialization_info(
            asset_record.asset_entry.asset_key
        )
        return planned_info.run_id if planned_info else None


def get_assets_latest_info(
    graphene_info: "ResolveInfo", step_keys_by_asset: Mapping[AssetKey, Sequence[str]]
) -> Sequence["GrapheneAssetLatestInfo"]:
    from dagster_graphql.schema.asset_graph import GrapheneAssetLatestInfo
    from dagster_graphql.schema.logs.events import GrapheneMaterializationEvent
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

    instance = graphene_info.context.instance

    asset_keys = list(step_keys_by_asset.keys())

    if not asset_keys:
        return []

    asset_nodes = {
        asset_key: graphene_info.context.asset_graph.get(asset_key) for asset_key in asset_keys
    }

    asset_records = AssetRecord.blocking_get_many(graphene_info.context, asset_keys)

    latest_materialization_by_asset = {
        asset_record.asset_entry.asset_key: (
            GrapheneMaterializationEvent(event=asset_record.asset_entry.last_materialization)
            if asset_record.asset_entry.last_materialization
            and asset_record.asset_entry.asset_key in step_keys_by_asset
            else None
        )
        for asset_record in asset_records
    }

    # Build a lookup table of asset keys to last materialization run IDs. We will filter these
    # run IDs out of the "in progress" run lists that are generated below since they have already
    # emitted an output for the run.
    latest_materialization_run_id_by_asset: dict[AssetKey, Optional[str]] = {
        asset_record.asset_entry.asset_key: (
            asset_record.asset_entry.last_materialization.run_id
            if asset_record.asset_entry.last_materialization
            and asset_record.asset_entry.asset_key in step_keys_by_asset
            else None
        )
        for asset_record in asset_records
    }

    latest_planned_run_ids_by_asset = {
        k: v
        for k, v in {
            asset_record.asset_entry.asset_key: _get_latest_planned_run_id(instance, asset_record)
            for asset_record in asset_records
        }.items()
        if v
    }

    run_records_by_run_id = {}

    run_ids = (
        list(set(latest_planned_run_ids_by_asset.values()))
        if latest_planned_run_ids_by_asset
        else []
    )
    if run_ids:
        run_records = instance.get_run_records(RunsFilter(run_ids=run_ids))
        for run_record in run_records:
            run_records_by_run_id[run_record.dagster_run.run_id] = run_record

    (
        in_progress_run_ids_by_asset,
        unstarted_run_ids_by_asset,
    ) = _get_in_progress_runs_for_assets(
        run_records_by_run_id,
        latest_materialization_run_id_by_asset,
        latest_planned_run_ids_by_asset,
    )

    from dagster_graphql.implementation.fetch_assets import get_unique_asset_id

    latest_infos = []
    for asset_key in step_keys_by_asset.keys():
        asset_node = asset_nodes[asset_key]
        if asset_node:
            handle = asset_node.resolve_to_singular_repo_scoped_node().repository_handle
            node_id = get_unique_asset_id(
                asset_key,
                handle.repository_name,
                handle.location_name,
            )
        else:
            node_id = get_unique_asset_id(asset_key)

        latest_infos.append(
            GrapheneAssetLatestInfo(
                id=node_id,
                assetKey=asset_key,
                latestMaterialization=latest_materialization_by_asset.get(asset_key),
                unstartedRunIds=list(unstarted_run_ids_by_asset.get(asset_key, [])),
                inProgressRunIds=list(in_progress_run_ids_by_asset.get(asset_key, [])),
                latestRun=(
                    GrapheneRun(run_records_by_run_id[latest_planned_run_ids_by_asset[asset_key]])
                    # Dagster UI error occurs if a run is terminated at the same time that this endpoint is
                    # called so we check to make sure the run ID exists in the run records.
                    if asset_key in latest_planned_run_ids_by_asset
                    and latest_planned_run_ids_by_asset[asset_key] in run_records_by_run_id
                    else None
                ),
            )
        )

    return latest_infos


def _get_in_progress_runs_for_assets(
    run_records_by_run_id: Mapping[str, RunRecord],
    latest_materialization_run_id_by_asset: dict[AssetKey, Optional[str]],
    latest_run_ids_by_asset: dict[AssetKey, str],
) -> tuple[Mapping[AssetKey, AbstractSet[str]], Mapping[AssetKey, AbstractSet[str]]]:
    in_progress_run_ids_by_asset = defaultdict(set)
    unstarted_run_ids_by_asset = defaultdict(set)

    for asset_key, run_id in latest_run_ids_by_asset.items():
        record = run_records_by_run_id.get(run_id)

        if not record:
            continue

        run = record.dagster_run
        if (
            run.status not in PENDING_STATUSES
            or latest_materialization_run_id_by_asset.get(asset_key) == run.run_id
        ):
            continue

        if run.status in IN_PROGRESS_STATUSES:
            in_progress_run_ids_by_asset[asset_key].add(record.dagster_run.run_id)
        else:
            unstarted_run_ids_by_asset[asset_key].add(record.dagster_run.run_id)

    return in_progress_run_ids_by_asset, unstarted_run_ids_by_asset


def get_runs_count(graphene_info: "ResolveInfo", filters: Optional[RunsFilter]) -> int:
    return graphene_info.context.instance.get_runs_count(filters)


def validate_pipeline_config(
    graphene_info: "ResolveInfo",
    selector: JobSubsetSelector,
    run_config: Union[str, Mapping[str, object]],
) -> "GraphenePipelineConfigValidationValid":
    from dagster_graphql.schema.pipelines.config import GraphenePipelineConfigValidationValid

    check.inst_param(selector, "selector", JobSubsetSelector)

    remote_job = get_remote_job_or_raise(graphene_info, selector)
    ensure_valid_config(remote_job, run_config)
    return GraphenePipelineConfigValidationValid(pipeline_name=remote_job.name)


def get_execution_plan(
    graphene_info: "ResolveInfo",
    selector: JobSubsetSelector,
    run_config: Mapping[str, Any],
) -> "GrapheneExecutionPlan":
    from dagster_graphql.schema.execution import GrapheneExecutionPlan

    check.inst_param(selector, "selector", JobSubsetSelector)

    remote_job = get_remote_job_or_raise(graphene_info, selector)
    ensure_valid_config(remote_job, run_config)
    return GrapheneExecutionPlan(
        graphene_info.context.get_execution_plan(
            remote_job=remote_job,
            run_config=run_config,
            step_keys_to_execute=None,
            known_state=None,
        )
    )


def get_stats(graphene_info: "ResolveInfo", run_id: str) -> "GrapheneRunStatsSnapshot":
    from dagster_graphql.schema.pipelines.pipeline_run_stats import GrapheneRunStatsSnapshot

    stats = graphene_info.context.instance.get_run_stats(run_id)
    stats.id = "stats-{run_id}"  # type: ignore  # (unused code path)
    return GrapheneRunStatsSnapshot(stats)


def get_step_stats(
    graphene_info: "ResolveInfo", run_id: str, step_keys: Optional[Sequence[str]] = None
) -> Sequence["GrapheneRunStepStats"]:
    from dagster_graphql.schema.logs.events import GrapheneRunStepStats

    step_stats = graphene_info.context.instance.get_run_step_stats(run_id, step_keys)
    return [GrapheneRunStepStats(stats) for stats in step_stats]


def get_logs_for_run(
    graphene_info: "ResolveInfo",
    run_id: str,
    cursor: Optional[str] = None,
    limit: Optional[int] = None,
) -> Union["GrapheneRunNotFoundError", "GrapheneEventConnection"]:
    from dagster_graphql.implementation.events import from_event_record
    from dagster_graphql.schema.errors import GrapheneRunNotFoundError
    from dagster_graphql.schema.pipelines.pipeline import GrapheneEventConnection

    instance = graphene_info.context.instance
    run = instance.get_run_by_id(run_id)
    if not run:
        return GrapheneRunNotFoundError(run_id)

    conn = instance.get_records_for_run(run_id, cursor=cursor, limit=limit)
    return GrapheneEventConnection(
        events=[from_event_record(record.event_log_entry, run.job_name) for record in conn.records],
        cursor=conn.cursor,
        hasMore=conn.has_more,
    )


@record
class RunsFeedCursor:
    """Three part cursor for paginating the Runs Feed. The run_cursor is the run_id of the oldest run that
    has been returned. The backfill_cursor is the id of the oldest backfill that has been returned. The
    timestamp is the timestamp of the oldest entry (run or backfill).

    If the run/backfill cursor is None, that means that no runs/backfills have been returned yet and querying
    should begin at the start of the table. Once all runs/backfills in the table have been returned, the
    corresponding cursor should still be set to the id of the last run/backfill returned.

    The timestamp is used for the following case. If a deployment has 20 runs and 0 backfills, and a query is
    made for 10 Runs Feed entries, the first 10 runs will be returned. At this time, the run_cursor will be an id,
    and the backfill_cursor will be None. Then a backfill is created. If a second query is made for 10 Runs Feed entries
    the newly created backfill will get included in the list, even though it should be included on the first page by time
    order. To prevent this, the timestamp is used to ensure that all returned entires are older than the entries on the
    previous page.
    """

    run_cursor: Optional[str]
    backfill_cursor: Optional[str]
    timestamp: Optional[float]

    def to_string(self) -> str:
        return f"{self.run_cursor if self.run_cursor else ''}{_DELIMITER}{self.backfill_cursor if self.backfill_cursor else ''}{_DELIMITER}{self.timestamp if self.timestamp else ''}"

    @staticmethod
    def from_string(serialized: Optional[str]):
        if serialized is None:
            return RunsFeedCursor(
                run_cursor=None,
                backfill_cursor=None,
                timestamp=None,
            )
        parts = serialized.split(_DELIMITER)
        if len(parts) != 3:
            raise DagsterInvariantViolationError(f"Invalid cursor for querying runs: {serialized}")

        return RunsFeedCursor(
            run_cursor=parts[0] if parts[0] else None,
            backfill_cursor=parts[1] if parts[1] else None,
            timestamp=float(parts[2]) if parts[2] else None,
        )


def _fetch_runs_not_in_backfill(
    instance: DagsterInstance,
    cursor: Optional[str],
    limit: int,
    filters: Optional[RunsFilter],
) -> Sequence[RunRecord]:
    """Fetches limit RunRecords that are not part of a backfill and were created before a given timestamp."""
    runs = []
    while len(runs) < limit:
        # fetch runs in a loop and discard runs that are part of a backfill until we have
        # limit runs to return or have reached the end of the runs table
        new_runs = instance.get_run_records(limit=limit, cursor=cursor, filters=filters)
        if len(new_runs) == 0:
            return runs
        cursor = new_runs[-1].dagster_run.run_id
        runs.extend([run for run in new_runs if run.dagster_run.tags.get(BACKFILL_ID_TAG) is None])

    return runs[:limit]


RUN_STATUS_TO_BULK_ACTION_STATUSES = {
    DagsterRunStatus.SUCCESS: [BulkActionStatus.COMPLETED_SUCCESS],
    DagsterRunStatus.FAILURE: [BulkActionStatus.FAILED, BulkActionStatus.COMPLETED_FAILED],
    DagsterRunStatus.CANCELED: [BulkActionStatus.CANCELED],
    DagsterRunStatus.CANCELING: [BulkActionStatus.CANCELING],
    DagsterRunStatus.STARTED: [BulkActionStatus.REQUESTED],
}


def _bulk_action_statuses_from_run_statuses(
    statuses: Sequence[DagsterRunStatus],
) -> Sequence[BulkActionStatus]:
    full_list = []
    for status in statuses:
        full_list.extend(RUN_STATUS_TO_BULK_ACTION_STATUSES.get(status, []))

    return full_list


def _filters_apply_to_backfills(filters: RunsFilter) -> bool:
    # the following filters do not apply to backfills, so skip fetching backfills if they are set
    if (
        (filters.run_ids is not None and len(filters.run_ids) > 0)
        or filters.updated_after is not None
        or filters.updated_before is not None
        or filters.snapshot_id is not None
    ):
        return False
    # if filtering by statuses and all are not valid backfill statuses, skip fetching backfills
    if filters.statuses and len(_bulk_action_statuses_from_run_statuses(filters.statuses)) == 0:
        return False

    return True


def _bulk_action_filters_from_run_filters(filters: RunsFilter) -> BulkActionsFilter:
    converted_statuses = (
        _bulk_action_statuses_from_run_statuses(filters.statuses) if filters.statuses else None
    )
    backfill_ids = None
    if filters.tags.get(BACKFILL_ID_TAG) is not None:
        backfill_ids = filters.tags[BACKFILL_ID_TAG]
        if isinstance(backfill_ids, str):
            backfill_ids = [backfill_ids]

    tags = (
        {key: value for key, value in filters.tags.items() if key != BACKFILL_ID_TAG}
        if filters.tags
        else None
    )
    return BulkActionsFilter(
        created_before=filters.created_before,
        created_after=filters.created_after,
        statuses=converted_statuses,
        job_name=filters.job_name,
        tags=tags,
        backfill_ids=backfill_ids,
    )


def _replace_created_before_with_cursor(
    filters: RunsFilter, created_before_cursor: Optional[datetime.datetime]
):
    """After the first page of results is returned, created_before_cursor will be less than
    filters.created_before. For pagination of results to work, we need to ensure that the
    created_before filter is set to the minimum of created_before_cursor and filters.created_before.
    """
    if filters.created_before and created_before_cursor:
        created_before = min(created_before_cursor, filters.created_before)
    elif created_before_cursor:
        created_before = created_before_cursor
    elif filters.created_before:
        created_before = filters.created_before
    else:  # no created_before should be applied, return filters as is
        return filters

    return copy(filters, created_before=created_before)


def get_runs_feed_entries(
    graphene_info: "ResolveInfo",
    limit: int,
    filters: Optional[RunsFilter],
    view: "GrapheneRunsFeedView",
    cursor: Optional[str] = None,
) -> "GrapheneRunsFeedConnection":
    """Returns a GrapheneRunsFeedConnection, which contains a merged list of backfills and
    single runs (runs that are not part of a backfill), the cursor to fetch the next page,
    and a boolean indicating if there are more results to fetch.

    Args:
        limit (int): max number of results to return
        cursor (Optional[str]): String that can be deserialized into a RunsFeedCursor. If None, indicates
            that querying should start at the beginning of the table for both runs and backfills.
        filters (Optional[RunsFilter]): Filters to apply to the runs. If None, no filters are applied.
        view (RunsFeedView): If True, include runs that are part of a backfill in the results and exclude backfill objects
    """
    from dagster_graphql.schema.backfill import GraphenePartitionBackfill
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
    from dagster_graphql.schema.runs_feed import GrapheneRunsFeedConnection, GrapheneRunsFeedView

    check.opt_str_param(cursor, "cursor")
    check.int_param(limit, "limit")
    check.opt_inst_param(filters, "filters", RunsFilter)

    instance = graphene_info.context.instance
    runs_feed_cursor = RunsFeedCursor.from_string(cursor)
    # In the default "ROOTS" run feed, we exclude runs that are part of backfills. If
    # the user chooses the "RUNS" view, we want to flatten backfills into their runs.
    exclude_subruns = view == GrapheneRunsFeedView.ROOTS

    # if using limit, fetch limit+1 of each type to know if there are more than limit remaining
    fetch_limit = limit + 1
    # filter out any backfills/runs that are newer than the cursor timestamp. See RunsFeedCursor docstring
    # for case when this is necessary
    created_before_cursor = (
        datetime_from_timestamp(runs_feed_cursor.timestamp) if runs_feed_cursor.timestamp else None
    )

    should_fetch_backfills = (
        view == GrapheneRunsFeedView.ROOTS or view == GrapheneRunsFeedView.BACKFILLS
    ) and (_filters_apply_to_backfills(filters) if filters else True)
    if filters:
        check.invariant(
            filters.exclude_subruns is None,
            "filters.exclude_subruns must be None when fetching the runs feed. Use include_runs_from_backfills instead.",
        )
        with disable_dagster_warnings():
            run_filters = copy(filters, exclude_subruns=exclude_subruns)
            run_filters = _replace_created_before_with_cursor(run_filters, created_before_cursor)
        backfill_filters = (
            _bulk_action_filters_from_run_filters(run_filters) if should_fetch_backfills else None
        )
    else:
        with disable_dagster_warnings():
            run_filters = RunsFilter(
                created_before=created_before_cursor, exclude_subruns=exclude_subruns
            )
        backfill_filters = BulkActionsFilter(created_before=created_before_cursor)

    if should_fetch_backfills:
        backfills = [
            GraphenePartitionBackfill(backfill)
            for backfill in instance.get_backfills(
                cursor=runs_feed_cursor.backfill_cursor,
                limit=fetch_limit,
                filters=backfill_filters,
            )
        ]
    else:
        backfills = []

    # if we are not showing runs within backfills and the backfill_id filter is set, we know
    # there will be no results, so we can skip fetching runs
    should_fetch_runs = (
        view == GrapheneRunsFeedView.ROOTS or view == GrapheneRunsFeedView.RUNS
    ) and not (exclude_subruns and run_filters.tags.get(BACKFILL_ID_TAG) is not None)
    if should_fetch_runs:
        runs = [
            GrapheneRun(run)
            for run in instance.get_run_records(
                limit=fetch_limit, cursor=runs_feed_cursor.run_cursor, filters=run_filters
            )
        ]
    else:
        runs = []

    # if we fetched limit+1 of either runs or backfills, we know there must be more results
    # to fetch on the next call since we will return limit results for this call. Additionally,
    # if we fetched more than limit of runs and backfill combined, we know there are more results
    has_more = (
        len(backfills) == fetch_limit
        or len(runs) == fetch_limit
        or len(backfills) + len(runs) > limit
    )

    all_entries = backfills + runs

    # order runs and backfills by create_time. typically we sort by storage id but that won't work here since
    # they are different tables
    all_entries = sorted(
        all_entries,
        key=lambda x: x.creation_timestamp,
        reverse=True,
    )

    to_return = all_entries[:limit]

    new_run_cursor = None
    new_backfill_cursor = None
    for entry in reversed(to_return):
        if new_run_cursor is not None and new_backfill_cursor is not None:
            break
        if new_backfill_cursor is None and isinstance(entry, GraphenePartitionBackfill):
            new_backfill_cursor = entry.id
        if new_run_cursor is None and isinstance(entry, GrapheneRun):
            new_run_cursor = entry.runId

    new_timestamp = to_return[-1].creation_timestamp if to_return else None

    # if either of the new cursors are None, replace with the cursor passed in so the next call doesn't
    # restart at the top the table.
    final_cursor = RunsFeedCursor(
        run_cursor=new_run_cursor if new_run_cursor else runs_feed_cursor.run_cursor,
        backfill_cursor=new_backfill_cursor
        if new_backfill_cursor
        else runs_feed_cursor.backfill_cursor,
        timestamp=new_timestamp if new_timestamp else runs_feed_cursor.timestamp,
    )

    return GrapheneRunsFeedConnection(
        results=to_return, cursor=final_cursor.to_string(), hasMore=has_more
    )


def get_runs_feed_count(
    graphene_info: "ResolveInfo", filters: Optional[RunsFilter], view: "GrapheneRunsFeedView"
) -> int:
    from dagster_graphql.schema.runs_feed import GrapheneRunsFeedView

    # In the default "ROOTS" run feed, we exclude runs that are part of backfills. If
    # the user chooses the "RUNS" view, we want to flatten backfills into their runs.
    exclude_subruns = view == GrapheneRunsFeedView.ROOTS
    should_fetch_runs = view == GrapheneRunsFeedView.RUNS or view == GrapheneRunsFeedView.ROOTS
    should_fetch_backfills = (
        view == GrapheneRunsFeedView.BACKFILLS or view == GrapheneRunsFeedView.ROOTS
    ) and (_filters_apply_to_backfills(filters) if filters else True)
    with disable_dagster_warnings():
        run_filters = (
            copy(filters, exclude_subruns=exclude_subruns)
            if filters
            else RunsFilter(exclude_subruns=exclude_subruns)
        )
    if should_fetch_backfills:
        backfill_filters = _bulk_action_filters_from_run_filters(run_filters)
        backfills_count = graphene_info.context.instance.get_backfills_count(backfill_filters)
    else:
        backfills_count = 0

    if should_fetch_runs:
        runs_count = graphene_info.context.instance.get_runs_count(run_filters)
    else:
        runs_count = 0

    return runs_count + backfills_count



import functools
import sys
from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from datetime import datetime
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

import dagster._check as check
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.remote_asset_graph import RemoteWorkspaceAssetGraph
from dagster._core.definitions.selector import GraphSelector, JobSubsetSelector
from dagster._core.execution.backfill import PartitionBackfill
from dagster._utils.caching_instance_queryer import CachingInstanceQueryer
from dagster._utils.error import serializable_error_info_from_exc_info
from typing_extensions import ParamSpec, TypeAlias

if TYPE_CHECKING:
    from dagster._core.workspace.context import BaseWorkspaceRequestContext

    from dagster_graphql.schema.errors import GrapheneError, GraphenePythonError
    from dagster_graphql.schema.util import ResolveInfo

P = ParamSpec("P")
T = TypeVar("T")

GrapheneResolverFn: TypeAlias = Callable[..., object]
T_Callable = TypeVar("T_Callable", bound=Callable)


def assert_permission_for_location(
    graphene_info: "ResolveInfo", permission: str, location_name: str
) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    context = cast("BaseWorkspaceRequestContext", graphene_info.context)
    if not context.has_permission_for_location(permission, location_name):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def require_permission_check(
    permission: str,
) -> Callable[[GrapheneResolverFn], GrapheneResolverFn]:
    def decorator(fn: GrapheneResolverFn) -> GrapheneResolverFn:
        @functools.wraps(fn)
        def _fn(self, graphene_info, *args: P.args, **kwargs: P.kwargs):
            result = fn(self, graphene_info, *args, **kwargs)

            if not graphene_info.context.was_permission_checked(permission):
                raise Exception(f"Permission {permission} was never checked during the request")

            return result

        return _fn

    return decorator


def check_permission(
    permission: str,
) -> Callable[[GrapheneResolverFn], GrapheneResolverFn]:
    def decorator(fn: GrapheneResolverFn) -> GrapheneResolverFn:
        @functools.wraps(fn)
        def _fn(self, graphene_info, *args: P.args, **kwargs: P.kwargs):
            assert_permission(graphene_info, permission)

            return fn(self, graphene_info, *args, **kwargs)

        return _fn

    return decorator


def assert_permission(graphene_info: "ResolveInfo", permission: str) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    context = cast("BaseWorkspaceRequestContext", graphene_info.context)
    if not context.has_permission(permission):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def has_permission_for_asset_graph(
    graphene_info: "ResolveInfo",
    asset_graph: RemoteWorkspaceAssetGraph,
    asset_selection: Optional[Sequence[AssetKey]],
    permission: str,
) -> bool:
    asset_keys = set(asset_selection or [])
    context = cast("BaseWorkspaceRequestContext", graphene_info.context)

    if asset_keys:
        location_names = set()
        for key in asset_keys:
            if not asset_graph.has(key):
                # If any of the asset keys don't map to a location (e.g. because they are no longer in the
                # graph) need deployment-wide permissions - no valid code location to check
                return context.has_permission(permission)
            node = asset_graph.get(key)
            location_names.add(
                node.resolve_to_singular_repo_scoped_node().repository_handle.location_name
            )
    else:
        location_names = set(
            handle.location_name for handle in asset_graph.repository_handles_by_key.values()
        )

    if not location_names:
        return context.has_permission(permission)
    else:
        return all(
            context.has_permission_for_location(permission, location_name)
            for location_name in location_names
        )


def assert_permission_for_asset_graph(
    graphene_info: "ResolveInfo",
    asset_graph: RemoteWorkspaceAssetGraph,
    asset_selection: Optional[Sequence[AssetKey]],
    permission: str,
) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    if not has_permission_for_asset_graph(graphene_info, asset_graph, asset_selection, permission):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def assert_valid_job_partition_backfill(
    graphene_info: "ResolveInfo",
    backfill: PartitionBackfill,
    partitions_def: PartitionsDefinition,
    dynamic_partitions_store: CachingInstanceQueryer,
    backfill_datetime: datetime,
) -> None:
    from dagster_graphql.schema.errors import GraphenePartitionKeysNotFoundError

    partition_names = backfill.get_partition_names(graphene_info.context)

    if not partition_names:
        return

    invalid_keys = set(partition_names) - set(
        partitions_def.get_partition_keys(backfill_datetime, dynamic_partitions_store)
    )

    if invalid_keys:
        raise UserFacingGraphQLError(GraphenePartitionKeysNotFoundError(invalid_keys))


def assert_valid_asset_partition_backfill(
    graphene_info: "ResolveInfo",
    backfill: PartitionBackfill,
    dynamic_partitions_store: CachingInstanceQueryer,
    backfill_datetime: datetime,
) -> None:
    from dagster_graphql.schema.errors import GraphenePartitionKeysNotFoundError

    asset_graph = graphene_info.context.asset_graph
    asset_backfill_data = backfill.asset_backfill_data

    if not asset_backfill_data:
        return

    partition_subset_by_asset_key = (
        asset_backfill_data.target_subset.partitions_subsets_by_asset_key
    )

    for asset_key, partition_subset in partition_subset_by_asset_key.items():
        partitions_def = asset_graph.get(asset_key).partitions_def

        if not partitions_def:
            continue

        invalid_keys = set(partition_subset.get_partition_keys()) - set(
            partitions_def.get_partition_keys(backfill_datetime, dynamic_partitions_store)
        )

        if invalid_keys:
            raise UserFacingGraphQLError(GraphenePartitionKeysNotFoundError(invalid_keys))


def _noop(_) -> None:
    pass


class ErrorCapture:
    @staticmethod
    def default_on_exception(
        exc_info: tuple[type[BaseException], BaseException, TracebackType],
    ) -> "GraphenePythonError":
        from dagster_graphql.schema.errors import GraphenePythonError

        # Transform exception in to PythonError to present to user
        return GraphenePythonError(serializable_error_info_from_exc_info(exc_info))

    # global behavior for how to handle unexpected exceptions
    on_exception = default_on_exception

    # context var for observing unexpected exceptions
    observer: ContextVar[Callable[[Exception], None]] = ContextVar(
        "error_capture_observer", default=_noop
    )

    @staticmethod
    @contextmanager
    def watch(fn: Callable[[Exception], None]) -> Iterator[None]:
        token = ErrorCapture.observer.set(fn)
        try:
            yield
        finally:
            ErrorCapture.observer.reset(token)


@overload
def capture_error(
    fn: Callable[P, T],
) -> Callable[P, T]: ...


@overload
def capture_error(  # pyright: ignore[reportOverlappingOverload]
    fn: Callable[P, Awaitable[T]],
) -> Callable[P, Awaitable[T]]: ...


def capture_error(
    fn: Union[Callable[P, T], Callable[P, Awaitable[T]]],
) -> Union[
    Callable[P, Union[T, "GrapheneError", "GraphenePythonError"]],
    Callable[P, Awaitable[Union[T, "GrapheneError", "GraphenePythonError"]]],
]:
    if iscoroutinefunction(fn):

        @functools.wraps(fn)
        async def _async_fn(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return await fn(*args, **kwargs)
            except UserFacingGraphQLError as de_exception:
                return de_exception.error
            except Exception as exc:
                ErrorCapture.observer.get()(exc)
                return ErrorCapture.on_exception(sys.exc_info())  # type: ignore

        return _async_fn

    else:

        @functools.wraps(fn)
        def _fn(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return cast("T", fn(*args, **kwargs))
            except UserFacingGraphQLError as de_exception:
                return de_exception.error
            except Exception as exc:
                ErrorCapture.observer.get()(exc)
                return ErrorCapture.on_exception(sys.exc_info())  # type: ignore

        return _fn


class UserFacingGraphQLError(Exception):
    # The `error` arg here should be a Graphene type implementing the interface `GrapheneError`, but
    # this is not trackable by the Python type system.
    def __init__(self, error: Any):
        self.error = error
        message = "[{cls}] {message}".format(
            cls=error.__class__.__name__,
            message=error.message if hasattr(error, "message") else None,
        )
        super().__init__(message)


def pipeline_selector_from_graphql(data: Mapping[str, Any]) -> JobSubsetSelector:
    asset_selection = cast("Optional[Iterable[dict[str, list[str]]]]", data.get("assetSelection"))
    asset_check_selection = cast(
        "Optional[Iterable[dict[str, Any]]]", data.get("assetCheckSelection")
    )
    return JobSubsetSelector(
        location_name=data["repositoryLocationName"],
        repository_name=data["repositoryName"],
        job_name=data.get("pipelineName") or data.get("jobName"),  # type: ignore
        op_selection=data.get("solidSelection"),
        asset_selection=(
            [AssetKey.from_graphql_input(asset_key) for asset_key in asset_selection]
            if asset_selection
            else None
        ),
        asset_check_selection=(
            [AssetCheckKey.from_graphql_input(asset_check) for asset_check in asset_check_selection]
            if asset_check_selection is not None
            else None
        ),
    )


def graph_selector_from_graphql(data: Mapping[str, Any]) -> GraphSelector:
    return GraphSelector(
        location_name=data["repositoryLocationName"],
        repository_name=data["repositoryName"],
        graph_name=data["graphName"],
    )


class ExecutionParams(
    NamedTuple(
        "_ExecutionParams",
        [
            ("selector", JobSubsetSelector),
            ("run_config", Mapping[str, object]),
            ("mode", Optional[str]),
            ("execution_metadata", "ExecutionMetadata"),
            ("step_keys", Optional[Sequence[str]]),
        ],
    )
):
    def __new__(
        cls,
        selector: JobSubsetSelector,
        run_config: Optional[Mapping[str, object]],
        mode: Optional[str],
        execution_metadata: "ExecutionMetadata",
        step_keys: Optional[Sequence[str]],
    ):
        check.opt_list_param(step_keys, "step_keys", of_type=str)

        return super().__new__(
            cls,
            selector=check.inst_param(selector, "selector", JobSubsetSelector),
            run_config=check.opt_mapping_param(run_config, "run_config", key_type=str),
            mode=check.opt_str_param(mode, "mode"),
            execution_metadata=check.inst_param(
                execution_metadata, "execution_metadata", ExecutionMetadata
            ),
            step_keys=step_keys,
        )

    def to_graphql_input(self) -> Mapping[str, Any]:
        return {
            "selector": self.selector.to_graphql_input(),
            "runConfigData": self.run_config,
            "mode": self.mode,
            "executionMetadata": self.execution_metadata.to_graphql_input(),
            "stepKeys": self.step_keys,
        }


class ExecutionMetadata(
    NamedTuple(
        "_ExecutionMetadata",
        [
            ("run_id", Optional[str]),
            ("tags", Mapping[str, str]),
            ("root_run_id", Optional[str]),
            ("parent_run_id", Optional[str]),
        ],
    )
):
    def __new__(
        cls,
        run_id: Optional[str],
        tags: Mapping[str, str],
        root_run_id: Optional[str] = None,
        parent_run_id: Optional[str] = None,
    ):
        return super().__new__(
            cls,
            check.opt_str_param(run_id, "run_id"),
            check.dict_param(tags, "tags", key_type=str, value_type=str),
            check.opt_str_param(root_run_id, "root_run_id"),
            check.opt_str_param(parent_run_id, "parent_run_id"),
        )

    def to_graphql_input(self) -> Mapping[str, Any]:
        return {
            "runId": self.run_id,
            "tags": [{"key": k, "value": v} for k, v in self.tags.items()],
            "rootRunId": self.root_run_id,
            "parentRunId": self.parent_run_id,
        }


def apply_cursor_limit_reverse(
    items: Sequence[str],
    cursor: Optional[str],
    limit: Optional[int],
    reverse: Optional[bool],
) -> Sequence[str]:
    start = 0
    end = len(items)
    index = 0

    if cursor:
        index = next(idx for (idx, item) in enumerate(items) if item == cursor)

        if reverse:
            end = index
        else:
            start = index + 1

    if limit:
        if reverse:
            start = end - limit
        else:
            end = start + limit

    return items[max(start, 0) : end]


BackfillParams: TypeAlias = Mapping[str, Any]
AssetBackfillPreviewParams: TypeAlias = Mapping[str, Any]



from typing import TYPE_CHECKING, Any, Optional

import dagster._check as check
import graphene
from dagster._core.remote_representation import RepresentedJob
from dagster._core.remote_representation.external_data import DEFAULT_MODE_NAME
from dagster._core.snap.snap_to_yaml import default_values_yaml_from_type_snap

from dagster_graphql.implementation.run_config_schema import resolve_is_run_config_valid
from dagster_graphql.implementation.utils import capture_error
from dagster_graphql.schema.config_types import GrapheneConfigType, to_config_type
from dagster_graphql.schema.errors import (
    GrapheneInvalidSubsetError,
    GrapheneModeNotFoundError,
    GraphenePipelineNotFoundError,
    GraphenePythonError,
)
from dagster_graphql.schema.pipelines.config_result import GraphenePipelineConfigValidationResult
from dagster_graphql.schema.runs import GrapheneRunConfigData, parse_run_config_input
from dagster_graphql.schema.util import ResolveInfo, non_null_list

if TYPE_CHECKING:
    from dagster._config.snap import ConfigSchemaSnapshot


class GrapheneRunConfigSchema(graphene.ObjectType):
    rootConfigType = graphene.Field(
        graphene.NonNull(GrapheneConfigType),
        description="""Fetch the root environment type. Concretely this is the type that
        is in scope at the root of configuration document for a particular execution selection.
        It is the type that is in scope initially with a blank config editor.""",
    )
    allConfigTypes = graphene.Field(
        non_null_list(GrapheneConfigType),
        description="""Fetch all the named config types that are in the schema. Useful
        for things like a type browser UI, or for fetching all the types are in the
        scope of a document so that the index can be built for the autocompleting editor.
    """,
    )

    isRunConfigValid = graphene.Field(
        graphene.NonNull(GraphenePipelineConfigValidationResult),
        runConfigData=graphene.Argument(GrapheneRunConfigData),
        description="""Parse a particular run config result. The return value
        either indicates that the validation succeeded by returning
        `PipelineConfigValidationValid` or that there are configuration errors
        by returning `RunConfigValidationInvalid' which containers a list errors
        so that can be rendered for the user""",
    )

    rootDefaultYaml = graphene.Field(
        graphene.NonNull(graphene.String),
        description="""The default configuration for this run in yaml. This is
        so that the client does not have to parse JSON client side and assemble
        it into a single yaml document.""",
    )

    class Meta:
        description = """The run config schema represents the all the config type
        information given a certain execution selection and mode of execution of that
        selection. All config interactions (e.g. checking config validity, fetching
        all config types, fetching in a particular config type) should be done
        through this type """
        name = "RunConfigSchema"

    def __init__(self, represented_job: RepresentedJob, mode: str):
        super().__init__()
        self._represented_job = check.inst_param(represented_job, "represented_job", RepresentedJob)
        self._mode = check.str_param(mode, "mode")

    def resolve_allConfigTypes(self, _graphene_info: ResolveInfo):
        return sorted(
            list(
                map(
                    lambda key: to_config_type(
                        self._represented_job.config_schema_snapshot.get_config_snap, key
                    ),
                    self._represented_job.config_schema_snapshot.all_config_keys,
                )
            ),
            key=lambda ct: ct.key,
        )

    def resolve_rootConfigType(self, _graphene_info: ResolveInfo):
        return to_config_type(
            self._represented_job.config_schema_snapshot.get_config_snap,
            self._represented_job.get_mode_def_snap(  # type: ignore  # (possible none)
                self._mode or DEFAULT_MODE_NAME
            ).root_config_key,
        )

    @capture_error
    def resolve_isRunConfigValid(
        self,
        graphene_info: ResolveInfo,
        runConfigData: Optional[Any] = None,  # custom scalar (GrapheneRunConfigData)
    ):
        return resolve_is_run_config_valid(
            graphene_info,
            self._represented_job,
            self._mode,
            parse_run_config_input(runConfigData or {}, raise_on_error=False),  # type: ignore
        )

    def resolve_rootDefaultYaml(self, _graphene_info) -> str:
        config_schema_snapshot: ConfigSchemaSnapshot = self._represented_job.config_schema_snapshot

        root_key = check.not_none(
            self._represented_job.get_mode_def_snap(self._mode).root_config_key
        )

        root_type = config_schema_snapshot.get_config_snap(root_key)

        return default_values_yaml_from_type_snap(config_schema_snapshot, root_type)


class GrapheneRunConfigSchemaOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneRunConfigSchema,
            GraphenePipelineNotFoundError,
            GrapheneInvalidSubsetError,
            GrapheneModeNotFoundError,
            GraphenePythonError,
        )
        name = "RunConfigSchemaOrError"



from collections.abc import Sequence
from typing import TYPE_CHECKING

from dagster._core.definitions.selector import RepositorySelector
from dagster._core.workspace.permissions import Permissions

from dagster_graphql.implementation.utils import (
    UserFacingGraphQLError,
    assert_permission_for_location,
)
from dagster_graphql.schema.errors import GrapheneDuplicateDynamicPartitionError

if TYPE_CHECKING:
    from dagster_graphql.schema.inputs import GrapheneRepositorySelector
    from dagster_graphql.schema.partition_sets import (
        GrapheneAddDynamicPartitionSuccess,
        GrapheneDeleteDynamicPartitionsSuccess,
    )


def _repository_contains_dynamic_partitions_def(
    graphene_info, repository_selector: RepositorySelector, partitions_def_name: str
) -> bool:
    from dagster._core.remote_representation.external_data import (
        DynamicPartitionsSnap,
        MultiPartitionsSnap,
        PartitionsSnap,
    )

    def _is_matching_partitions_def(partitions_snap: PartitionsSnap):
        if isinstance(partitions_snap, DynamicPartitionsSnap):
            return partitions_snap.name == partitions_def_name
        if isinstance(partitions_snap, MultiPartitionsSnap):
            return any(
                [
                    _is_matching_partitions_def(dimension.partitions)
                    for dimension in partitions_snap.partition_dimensions
                ]
            )
        return False

    if graphene_info.context.has_code_location(repository_selector.location_name):
        repo_loc = graphene_info.context.get_code_location(repository_selector.location_name)
        if repo_loc.has_repository(repository_selector.repository_name):
            repository = repo_loc.get_repository(repository_selector.repository_name)
            found_partitions_defs = [
                asset_node_snap.partitions
                for asset_node_snap in repository.repository_snap.asset_nodes
                if asset_node_snap.partitions
            ]
            return any(
                [
                    _is_matching_partitions_def(partitions_def)
                    for partitions_def in found_partitions_defs
                ]
            )
    return False


def add_dynamic_partition(
    graphene_info,
    repository_selector: "GrapheneRepositorySelector",
    partitions_def_name: str,
    partition_key: str,
) -> "GrapheneAddDynamicPartitionSuccess":
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError
    from dagster_graphql.schema.partition_sets import GrapheneAddDynamicPartitionSuccess

    unpacked_repository_selector = RepositorySelector.from_graphql_input(repository_selector)

    assert_permission_for_location(
        graphene_info,
        Permissions.EDIT_DYNAMIC_PARTITIONS,
        unpacked_repository_selector.location_name,
    )

    if not _repository_contains_dynamic_partitions_def(
        graphene_info, unpacked_repository_selector, partitions_def_name
    ):
        raise UserFacingGraphQLError(
            GrapheneUnauthorizedError(
                message=(
                    "The repository does not contain a dynamic partitions definition with the given"
                    " name."
                )
            )
        )

    if graphene_info.context.instance.has_dynamic_partition(partitions_def_name, partition_key):
        raise UserFacingGraphQLError(
            GrapheneDuplicateDynamicPartitionError(partitions_def_name, partition_key)
        )

    graphene_info.context.instance.add_dynamic_partitions(partitions_def_name, [partition_key])
    return GrapheneAddDynamicPartitionSuccess(
        partitionsDefName=partitions_def_name, partitionKey=partition_key
    )


def delete_dynamic_partitions(
    graphene_info,
    repository_selector: "GrapheneRepositorySelector",
    partitions_def_name: str,
    partition_keys: Sequence[str],
) -> "GrapheneDeleteDynamicPartitionsSuccess":
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError
    from dagster_graphql.schema.partition_sets import GrapheneDeleteDynamicPartitionsSuccess

    unpacked_repository_selector = RepositorySelector.from_graphql_input(repository_selector)

    assert_permission_for_location(
        graphene_info,
        Permissions.EDIT_DYNAMIC_PARTITIONS,
        unpacked_repository_selector.location_name,
    )

    if not _repository_contains_dynamic_partitions_def(
        graphene_info, unpacked_repository_selector, partitions_def_name
    ):
        raise UserFacingGraphQLError(
            GrapheneUnauthorizedError(
                message=(
                    "The repository does not contain a dynamic partitions definition with the given"
                    " name."
                )
            )
        )

    for partition_key in partition_keys:
        graphene_info.context.instance.delete_dynamic_partition(partitions_def_name, partition_key)

    return GrapheneDeleteDynamicPartitionsSuccess(partitionsDefName=partitions_def_name)



from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.freshness import FreshnessState
from dagster._core.storage.event_log.base import AssetRecord
from dagster._streamline.asset_check_health import AssetCheckHealthState
from dagster._streamline.asset_freshness_health import AssetFreshnessHealthState
from dagster._streamline.asset_health import AssetHealthStatus
from dagster._streamline.asset_materialization_health import AssetMaterializationHealthState

if TYPE_CHECKING:
    from dagster_graphql.schema.asset_health import (
        GrapheneAssetHealthCheckMeta,
        GrapheneAssetHealthFreshnessMeta,
        GrapheneAssetHealthMaterializationMeta,
    )
    from dagster_graphql.schema.util import ResolveInfo


async def get_asset_check_status_and_metadata(
    graphene_info: "ResolveInfo",
    asset_key: AssetKey,
) -> tuple[str, Optional["GrapheneAssetHealthCheckMeta"]]:
    """Converts an AssetCheckHealthState object to a GrapheneAssetHealthStatus and the metadata
    needed to power the UIs.
    """
    from dagster_graphql.schema.asset_health import (
        GrapheneAssetHealthCheckDegradedMeta,
        GrapheneAssetHealthCheckUnknownMeta,
        GrapheneAssetHealthCheckWarningMeta,
        GrapheneAssetHealthStatus,
    )

    if graphene_info.context.instance.streamline_read_asset_health_supported():
        asset_check_health_state = (
            graphene_info.context.instance.get_asset_check_health_state_for_asset(asset_key)
        )
        if asset_check_health_state is None:
            # asset_check_health_state_for is only None if no are defined on the asset
            return GrapheneAssetHealthStatus.NOT_APPLICABLE, None
    else:
        remote_check_nodes = graphene_info.context.asset_graph.get_checks_for_asset(asset_key)
        asset_check_health_state = await AssetCheckHealthState.compute_for_asset_checks(
            {remote_check_node.asset_check.key for remote_check_node in remote_check_nodes},
            graphene_info.context,
        )

    if asset_check_health_state.health_status == AssetHealthStatus.HEALTHY:
        return GrapheneAssetHealthStatus.HEALTHY, None
    if asset_check_health_state.health_status == AssetHealthStatus.WARNING:
        return (
            GrapheneAssetHealthStatus.WARNING,
            GrapheneAssetHealthCheckWarningMeta(
                numWarningChecks=len(asset_check_health_state.warning_checks),
                totalNumChecks=len(asset_check_health_state.all_checks),
            ),
        )
    if asset_check_health_state.health_status == AssetHealthStatus.DEGRADED:
        return (
            GrapheneAssetHealthStatus.DEGRADED,
            GrapheneAssetHealthCheckDegradedMeta(
                numFailedChecks=len(asset_check_health_state.failing_checks),
                numWarningChecks=len(asset_check_health_state.warning_checks),
                totalNumChecks=len(asset_check_health_state.all_checks),
            ),
        )
    if asset_check_health_state.health_status == AssetHealthStatus.UNKNOWN:
        return (
            GrapheneAssetHealthStatus.UNKNOWN,
            GrapheneAssetHealthCheckUnknownMeta(
                numNotExecutedChecks=len(asset_check_health_state.all_checks)
                - len(asset_check_health_state.passing_checks)
                - len(asset_check_health_state.failing_checks)
                - len(asset_check_health_state.warning_checks),
                totalNumChecks=len(asset_check_health_state.all_checks),
            ),
        )
    elif asset_check_health_state.health_status == AssetHealthStatus.NOT_APPLICABLE:
        return GrapheneAssetHealthStatus.NOT_APPLICABLE, None
    else:
        check.failed(
            f"Unexpected asset check health status: {asset_check_health_state.health_status}"
        )


async def get_freshness_status_and_metadata(
    graphene_info: "ResolveInfo", asset_key: AssetKey
) -> tuple[str, Optional["GrapheneAssetHealthFreshnessMeta"]]:
    """Gets an AssetFreshnessHealthState object for an asset, either via streamline or by computing
    it based on the state of the DB. Then converts it to a GrapheneAssetHealthStatus and the metadata
    needed to power the UIs. Metadata is fetched from the AssetLatestMaterializationState object, again
    either via streamline or by computing it based on the state of the DB.
    """
    from dagster_graphql.schema.asset_health import (
        GrapheneAssetHealthFreshnessMeta,
        GrapheneAssetHealthStatus,
    )

    if graphene_info.context.instance.streamline_read_asset_health_supported():
        asset_freshness_health_state = (
            graphene_info.context.instance.get_asset_freshness_health_state_for_asset(asset_key)
        )
        if asset_freshness_health_state is None:
            # if the freshness state is None, it means that the asset hasn't been processed by streamline
            # yet. Return UNKNOWN and rely on the freshness daemon to update the state on the next iteration.
            return GrapheneAssetHealthStatus.UNKNOWN, None
    else:
        if graphene_info.context.asset_graph.get(asset_key).internal_freshness_policy is None:
            return GrapheneAssetHealthStatus.NOT_APPLICABLE, None
        asset_freshness_health_state = AssetFreshnessHealthState.compute_for_asset(
            asset_key,
            graphene_info.context,
        )

    asset_record = await AssetRecord.gen(graphene_info.context, asset_key)
    materialization_timestamp = (
        asset_record.asset_entry.last_materialization.timestamp
        if asset_record
        and asset_record.asset_entry
        and asset_record.asset_entry.last_materialization
        else None
    )

    if asset_freshness_health_state.freshness_state == FreshnessState.PASS:
        return GrapheneAssetHealthStatus.HEALTHY, GrapheneAssetHealthFreshnessMeta(
            lastMaterializedTimestamp=materialization_timestamp,
        )
    if asset_freshness_health_state.freshness_state == FreshnessState.WARN:
        return GrapheneAssetHealthStatus.WARNING, GrapheneAssetHealthFreshnessMeta(
            lastMaterializedTimestamp=materialization_timestamp,
        )
    if asset_freshness_health_state.freshness_state == FreshnessState.FAIL:
        return GrapheneAssetHealthStatus.DEGRADED, GrapheneAssetHealthFreshnessMeta(
            lastMaterializedTimestamp=materialization_timestamp,
        )
    elif asset_freshness_health_state.freshness_state == FreshnessState.UNKNOWN:
        return GrapheneAssetHealthStatus.UNKNOWN, None
    elif asset_freshness_health_state.freshness_state == FreshnessState.NOT_APPLICABLE:
        return GrapheneAssetHealthStatus.NOT_APPLICABLE, None

    else:
        check.failed(f"Unexpected freshness state: {asset_freshness_health_state.freshness_state}")


async def get_materialization_status_and_metadata(
    graphene_info: "ResolveInfo", asset_key: AssetKey
) -> tuple[str, Optional["GrapheneAssetHealthMaterializationMeta"]]:
    """Gets an AssetMaterializationHealthState object for an asset, either via streamline or by computing
    it based on the state of the DB. Then converts it to a GrapheneAssetHealthStatus and the metadata
    needed to power the UIs. Metadata is fetched from the AssetLatestMaterializationState object, again
    either via streamline or by computing it based on the state of the DB.
    """
    from dagster_graphql.schema.asset_health import (
        GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta,
        GrapheneAssetHealthMaterializationDegradedPartitionedMeta,
        GrapheneAssetHealthMaterializationHealthyPartitionedMeta,
        GrapheneAssetHealthStatus,
    )

    asset_materialization_health_state = None
    if graphene_info.context.instance.streamline_read_asset_health_supported():
        asset_materialization_health_state = (
            graphene_info.context.instance.get_asset_materialization_health_state_for_asset(
                asset_key
            )
        )
    # captures streamline disabled or consumer state doesn't exist (because it's an observable source asset)
    if asset_materialization_health_state is None:
        node_snap = graphene_info.context.asset_graph.get(asset_key)
        if node_snap.is_observable and not node_snap.is_materializable:  # observable source asset
            # get the asset record to see if there is an observation event
            asset_record = await AssetRecord.gen(graphene_info.context, asset_key)
            if asset_record and asset_record.asset_entry.last_observation is not None:
                return GrapheneAssetHealthStatus.HEALTHY, None
            return GrapheneAssetHealthStatus.UNKNOWN, None

        asset_materialization_health_state = (
            await AssetMaterializationHealthState.compute_for_asset(
                asset_key,
                node_snap.partitions_def,
                graphene_info.context,
            )
        )

    if asset_materialization_health_state.health_status == AssetHealthStatus.HEALTHY:
        num_missing = 0
        total_num_partitions = 0
        if asset_materialization_health_state.partitions_def is not None:
            total_num_partitions = (
                asset_materialization_health_state.partitions_def.get_num_partitions(
                    dynamic_partitions_store=graphene_info.context.instance
                )
            )
            # asset is health, so no partitions are failed
            num_materialized = len(
                asset_materialization_health_state.materialized_subset.subset_value
            )
            num_missing = total_num_partitions - num_materialized
        if num_missing > 0 and total_num_partitions > 0:
            meta = GrapheneAssetHealthMaterializationHealthyPartitionedMeta(
                numMissingPartitions=num_missing,
                totalNumPartitions=total_num_partitions,
            )
        else:
            # captures the case when asset is not partitioned, or the asset is partitioned and all partitions are materialized
            meta = None
        return GrapheneAssetHealthStatus.HEALTHY, meta
    elif asset_materialization_health_state.health_status == AssetHealthStatus.DEGRADED:
        if asset_materialization_health_state.partitions_def is not None:
            total_num_partitions = (
                asset_materialization_health_state.partitions_def.get_num_partitions(
                    dynamic_partitions_store=graphene_info.context.instance
                )
            )
            num_failed = len(asset_materialization_health_state.failed_subset.subset_value)
            num_materialized = len(
                asset_materialization_health_state.currently_materialized_subset.subset_value
            )
            num_missing = total_num_partitions - num_materialized - num_failed
            meta = GrapheneAssetHealthMaterializationDegradedPartitionedMeta(
                numFailedPartitions=num_failed,
                numMissingPartitions=num_missing,
                totalNumPartitions=total_num_partitions,
            )
        else:
            meta = GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta(
                failedRunId=asset_materialization_health_state.latest_terminal_run_id,
            )
        return GrapheneAssetHealthStatus.DEGRADED, meta
    elif asset_materialization_health_state.health_status == AssetHealthStatus.UNKNOWN:
        return GrapheneAssetHealthStatus.UNKNOWN, None
    else:
        check.failed(
            f"Unexpected materialization health status: {asset_materialization_health_state.health_status}"
        )



from collections.abc import Sequence
from typing import AbstractSet, Optional, cast  # noqa: UP035

import dagster._check as check
import graphene
from dagster import MultiPartitionsDefinition
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.errors import DagsterUserCodeProcessError
from dagster._core.remote_representation import RemoteJob, RemotePartitionSet, RepositoryHandle
from dagster._core.remote_representation.external_data import (
    DynamicPartitionsSnap,
    MultiPartitionsSnap,
    PartitionExecutionErrorSnap,
    PartitionsSnap,
    StaticPartitionsSnap,
    TimeWindowPartitionsSnap,
    job_name_for_partition_set_snap_name,
)
from dagster._core.storage.dagster_run import RunsFilter
from dagster._core.storage.tags import PARTITION_NAME_TAG, PARTITION_SET_TAG
from dagster._utils.merger import merge_dicts

from dagster_graphql.implementation.fetch_partition_sets import (
    get_partition_by_name,
    get_partition_config,
    get_partition_set_partition_runs,
    get_partition_set_partition_statuses,
    get_partition_tags,
    get_partitions,
)
from dagster_graphql.implementation.fetch_runs import get_runs
from dagster_graphql.implementation.utils import capture_error
from dagster_graphql.schema.backfill import GraphenePartitionBackfill
from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.errors import (
    GrapheneDuplicateDynamicPartitionError,
    GraphenePartitionSetNotFoundError,
    GraphenePipelineNotFoundError,
    GraphenePythonError,
    GrapheneUnauthorizedError,
)
from dagster_graphql.schema.inputs import GrapheneRunsFilter
from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
from dagster_graphql.schema.pipelines.status import GrapheneRunStatus
from dagster_graphql.schema.repository_origin import GrapheneRepositoryOrigin
from dagster_graphql.schema.tags import GraphenePipelineTag
from dagster_graphql.schema.util import ResolveInfo, non_null_list


class GrapheneAddDynamicPartitionSuccess(graphene.ObjectType):
    partitionsDefName = graphene.NonNull(graphene.String)
    partitionKey = graphene.NonNull(graphene.String)

    class Meta:
        name = "AddDynamicPartitionSuccess"


class GrapheneAddDynamicPartitionResult(graphene.Union):
    class Meta:
        types = (
            GrapheneAddDynamicPartitionSuccess,
            GrapheneUnauthorizedError,
            GraphenePythonError,
            GrapheneDuplicateDynamicPartitionError,
        )
        name = "AddDynamicPartitionResult"


class GrapheneDeleteDynamicPartitionsSuccess(graphene.ObjectType):
    partitionsDefName = graphene.NonNull(graphene.String)

    class Meta:
        name = "DeleteDynamicPartitionsSuccess"


class GrapheneDeleteDynamicPartitionsResult(graphene.Union):
    class Meta:
        types = (
            GrapheneDeleteDynamicPartitionsSuccess,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "DeleteDynamicPartitionsResult"


class GraphenePartitionTags(graphene.ObjectType):
    results = non_null_list(GraphenePipelineTag)

    class Meta:
        name = "PartitionTags"


class GraphenePartitionRunConfig(graphene.ObjectType):
    yaml = graphene.NonNull(graphene.String)

    class Meta:
        name = "PartitionRunConfig"


class GraphenePartitionRunConfigOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitionRunConfig, GraphenePythonError)
        name = "PartitionRunConfigOrError"


class GraphenePartitionStatus(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    partitionName = graphene.NonNull(graphene.String)
    runId = graphene.Field(graphene.String)
    runStatus = graphene.Field(GrapheneRunStatus)
    runDuration = graphene.Field(graphene.Float)

    class Meta:
        name = "PartitionStatus"


class GraphenePartitionRun(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    partitionName = graphene.NonNull(graphene.String)
    run = graphene.Field(GrapheneRun)

    class Meta:
        name = "PartitionRun"


class GraphenePartitionStatusCounts(graphene.ObjectType):
    runStatus = graphene.NonNull(GrapheneRunStatus)
    count = graphene.NonNull(graphene.Int)

    class Meta:
        name = "PartitionStatusCounts"


class GraphenePartitionStatuses(graphene.ObjectType):
    results = non_null_list(GraphenePartitionStatus)

    class Meta:
        name = "PartitionStatuses"


class GraphenePartitionStatusesOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitionStatuses, GraphenePythonError)
        name = "PartitionStatusesOrError"


class GrapheneAssetPartitionsStatusCounts(graphene.ObjectType):
    class Meta:
        name = "AssetPartitionsStatusCounts"

    assetKey = graphene.NonNull(GrapheneAssetKey)
    numPartitionsTargeted = graphene.NonNull(graphene.Int)
    numPartitionsInProgress = graphene.NonNull(graphene.Int)
    numPartitionsMaterialized = graphene.NonNull(graphene.Int)
    numPartitionsFailed = graphene.NonNull(graphene.Int)


class GrapheneUnpartitionedAssetStatus(graphene.ObjectType):
    class Meta:
        name = "UnpartitionedAssetStatus"

    assetKey = graphene.NonNull(GrapheneAssetKey)
    inProgress = graphene.NonNull(graphene.Boolean)
    materialized = graphene.NonNull(graphene.Boolean)
    failed = graphene.NonNull(graphene.Boolean)


class GrapheneAssetBackfillStatus(graphene.Union):
    class Meta:
        types = (GrapheneAssetPartitionsStatusCounts, GrapheneUnpartitionedAssetStatus)
        name = "AssetBackfillStatus"


class GraphenePartitionKeyRange(graphene.ObjectType):
    class Meta:
        name = "PartitionKeyRange"

    start = graphene.NonNull(graphene.String)
    end = graphene.NonNull(graphene.String)


class GraphenePartitionTagsOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitionTags, GraphenePythonError)
        name = "PartitionTagsOrError"


class GrapheneJobSelectionPartition(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    job_name = graphene.NonNull(graphene.String)
    runConfigOrError = graphene.NonNull(GraphenePartitionRunConfigOrError)
    tagsOrError = graphene.NonNull(GraphenePartitionTagsOrError)

    class Meta:
        name = "PartitionTagsAndConfig"

    def __init__(
        self,
        remote_job: RemoteJob,
        partition_name: str,
        selected_asset_keys: Optional[AbstractSet[AssetKey]],
    ):
        self._remote_job = remote_job
        self._partition_name = partition_name
        self._selected_asset_keys = selected_asset_keys

        super().__init__(name=partition_name, job_name=remote_job.name)

    @capture_error
    def resolve_runConfigOrError(self, graphene_info: ResolveInfo) -> GraphenePartitionRunConfig:
        return get_partition_config(
            graphene_info,
            self._remote_job.repository_handle,
            self._remote_job.name,
            self._partition_name,
            selected_asset_keys=self._selected_asset_keys,
        )

    @capture_error
    def resolve_tagsOrError(self, graphene_info: ResolveInfo) -> GraphenePartitionTags:
        return get_partition_tags(
            graphene_info,
            self._remote_job.repository_handle,
            self._remote_job.name,
            self._partition_name,
            selected_asset_keys=self._selected_asset_keys,
        )


class GraphenePartition(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    partition_set_name = graphene.NonNull(graphene.String)
    solid_selection = graphene.List(graphene.NonNull(graphene.String))
    mode = graphene.NonNull(graphene.String)
    runConfigOrError = graphene.NonNull(GraphenePartitionRunConfigOrError)
    tagsOrError = graphene.NonNull(GraphenePartitionTagsOrError)
    runs = graphene.Field(
        non_null_list(GrapheneRun),
        filter=graphene.Argument(GrapheneRunsFilter),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )
    status = graphene.Field(GrapheneRunStatus)

    class Meta:
        name = "Partition"

    def __init__(
        self,
        repository_handle: RepositoryHandle,
        remote_partition_set: RemotePartitionSet,
        partition_name: str,
    ):
        self._repository_handle = check.inst_param(
            repository_handle, "repository_handle", RepositoryHandle
        )
        self._remote_partition_set = check.inst_param(
            remote_partition_set, "remote_partition_set", RemotePartitionSet
        )
        self._partition_name = check.str_param(partition_name, "partition_name")

        super().__init__(
            name=partition_name,
            partition_set_name=remote_partition_set.name,
            solid_selection=remote_partition_set.op_selection,
            mode=remote_partition_set.mode,
        )

    @capture_error
    def resolve_runConfigOrError(self, graphene_info: ResolveInfo):
        return get_partition_config(
            graphene_info,
            self._repository_handle,
            job_name_for_partition_set_snap_name(self._remote_partition_set.name),
            self._partition_name,
            selected_asset_keys=None,
        )

    @capture_error
    def resolve_tagsOrError(self, graphene_info: ResolveInfo):
        return get_partition_tags(
            graphene_info,
            self._repository_handle,
            job_name_for_partition_set_snap_name(self._remote_partition_set.name),
            self._partition_name,
            selected_asset_keys=None,
        )

    def resolve_runs(
        self,
        graphene_info: ResolveInfo,
        filter: Optional[GrapheneRunsFilter] = None,  # noqa: A002
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ):
        partition_tags = {
            PARTITION_SET_TAG: self._remote_partition_set.name,
            PARTITION_NAME_TAG: self._partition_name,
        }
        if filter is not None:
            selector = filter.to_selector()
            runs_filter = RunsFilter(
                run_ids=selector.run_ids,
                job_name=selector.job_name,
                statuses=selector.statuses,
                tags=merge_dicts(selector.tags, partition_tags),
            )
        else:
            runs_filter = RunsFilter(tags=partition_tags)

        return get_runs(graphene_info, runs_filter, cursor=cursor, limit=limit)


class GraphenePartitions(graphene.ObjectType):
    results = non_null_list(GraphenePartition)

    class Meta:
        name = "Partitions"


class GraphenePartitionsOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitions, GraphenePythonError)
        name = "PartitionsOrError"


class GraphenePartitionSet(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    name = graphene.NonNull(graphene.String)
    pipeline_name = graphene.NonNull(graphene.String)
    solid_selection = graphene.List(graphene.NonNull(graphene.String))
    mode = graphene.NonNull(graphene.String)
    partitionsOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionsOrError),
        cursor=graphene.String(),
        limit=graphene.Int(),
        reverse=graphene.Boolean(),
    )
    partition = graphene.Field(GraphenePartition, partition_name=graphene.NonNull(graphene.String))
    partitionStatusesOrError = graphene.NonNull(GraphenePartitionStatusesOrError)
    partitionRuns = non_null_list(GraphenePartitionRun)
    repositoryOrigin = graphene.NonNull(GrapheneRepositoryOrigin)
    backfills = graphene.Field(
        non_null_list(GraphenePartitionBackfill),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )

    class Meta:
        name = "PartitionSet"

    def __init__(
        self,
        repository_handle: RepositoryHandle,
        remote_partition_set: RemotePartitionSet,
    ):
        self._repository_handle = check.inst_param(
            repository_handle, "repository_handle", RepositoryHandle
        )
        self._remote_partition_set = check.inst_param(
            remote_partition_set, "remote_partition_set", RemotePartitionSet
        )
        self._partition_names = None

        super().__init__(
            name=remote_partition_set.name,
            pipeline_name=remote_partition_set.job_name,
            solid_selection=remote_partition_set.op_selection,
            mode=remote_partition_set.mode,
        )

    def _get_partition_names(self, graphene_info: ResolveInfo) -> Sequence[str]:
        if self._partition_names is None:
            result = graphene_info.context.get_partition_names(
                repository_handle=self._repository_handle,
                job_name=self._remote_partition_set.job_name,
                instance=graphene_info.context.instance,
                selected_asset_keys=None,
            )
            if isinstance(result, PartitionExecutionErrorSnap):
                raise DagsterUserCodeProcessError.from_error_info(result.error)

            self._partition_names = result.partition_names

        return self._partition_names

    def resolve_id(self, _graphene_info: ResolveInfo):
        return self._remote_partition_set.get_remote_origin_id()

    @capture_error
    def resolve_partitionsOrError(
        self,
        graphene_info: ResolveInfo,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        reverse: Optional[bool] = None,
    ):
        return get_partitions(
            self._repository_handle,
            self._remote_partition_set,
            self._get_partition_names(graphene_info),
            cursor=cursor,
            limit=limit,
            reverse=reverse or False,
        )

    def resolve_partition(self, graphene_info: ResolveInfo, partition_name: str):
        return get_partition_by_name(
            graphene_info,
            self._repository_handle,
            self._remote_partition_set,
            partition_name,
        )

    def resolve_partitionRuns(self, graphene_info: ResolveInfo):
        return get_partition_set_partition_runs(
            graphene_info,
            self._remote_partition_set,
            self._get_partition_names(graphene_info),
        )

    @capture_error
    def resolve_partitionStatusesOrError(self, graphene_info: ResolveInfo):
        return get_partition_set_partition_statuses(
            graphene_info,
            self._remote_partition_set,
            self._get_partition_names(graphene_info),
        )

    def resolve_repositoryOrigin(self, _):
        origin = self._remote_partition_set.get_remote_origin().repository_origin
        return GrapheneRepositoryOrigin(origin)

    def resolve_backfills(
        self, graphene_info: ResolveInfo, cursor: Optional[str] = None, limit: Optional[int] = None
    ):
        matching = [
            backfill
            for backfill in graphene_info.context.instance.get_backfills(
                cursor=cursor,
            )
            if backfill.partition_set_origin
            and backfill.partition_set_origin.partition_set_name == self._remote_partition_set.name
            and backfill.partition_set_origin.repository_origin.repository_name
            == self._repository_handle.repository_name
        ]
        return [GraphenePartitionBackfill(backfill) for backfill in matching[:limit]]


class GraphenePartitionSetOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitionSet, GraphenePartitionSetNotFoundError, GraphenePythonError)
        name = "PartitionSetOrError"


class GraphenePartitionSets(graphene.ObjectType):
    results = non_null_list(GraphenePartitionSet)

    class Meta:
        name = "PartitionSets"


class GraphenePartitionSetsOrError(graphene.Union):
    class Meta:
        types = (GraphenePartitionSets, GraphenePipelineNotFoundError, GraphenePythonError)
        name = "PartitionSetsOrError"


class GraphenePartitionDefinitionType(graphene.Enum):
    TIME_WINDOW = "TIME_WINDOW"
    STATIC = "STATIC"
    MULTIPARTITIONED = "MULTIPARTITIONED"
    DYNAMIC = "DYNAMIC"

    class Meta:
        name = "PartitionDefinitionType"

    @classmethod
    def from_partition_def_data(cls, partition_def_data):
        check.inst_param(partition_def_data, "partition_def_data", PartitionsSnap)
        if isinstance(partition_def_data, StaticPartitionsSnap):
            return GraphenePartitionDefinitionType.STATIC
        elif isinstance(partition_def_data, TimeWindowPartitionsSnap):
            return GraphenePartitionDefinitionType.TIME_WINDOW
        elif isinstance(partition_def_data, MultiPartitionsSnap):
            return GraphenePartitionDefinitionType.MULTIPARTITIONED
        elif isinstance(partition_def_data, DynamicPartitionsSnap):
            return GraphenePartitionDefinitionType.DYNAMIC
        else:
            check.failed(
                f"Invalid external partitions definition data type: {type(partition_def_data)}"
            )


class GrapheneDimensionDefinitionType(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    description = graphene.NonNull(graphene.String)
    type = graphene.NonNull(GraphenePartitionDefinitionType)
    isPrimaryDimension = graphene.NonNull(graphene.Boolean)
    dynamicPartitionsDefinitionName = graphene.Field(graphene.String)

    class Meta:
        name = "DimensionDefinitionType"


class GraphenePartitionDefinition(graphene.ObjectType):
    description = graphene.NonNull(graphene.String)
    type = graphene.NonNull(GraphenePartitionDefinitionType)
    dimensionTypes = non_null_list(GrapheneDimensionDefinitionType)
    name = graphene.Field(graphene.String)
    fmt = graphene.Field(graphene.String)

    class Meta:
        name = "PartitionDefinition"

    def resolve_dimensionTypes(self, _graphene_info):
        partition_def_data = self._partition_def_data
        return (
            [
                GrapheneDimensionDefinitionType(
                    name=dim.name,
                    description=str(dim.partitions.get_partitions_definition()),
                    type=GraphenePartitionDefinitionType.from_partition_def_data(dim.partitions),
                    isPrimaryDimension=dim.name
                    == cast(
                        "MultiPartitionsDefinition", partition_def_data.get_partitions_definition()
                    ).primary_dimension.name,
                    dynamicPartitionsDefinitionName=(
                        dim.partitions.name
                        if isinstance(
                            dim.partitions,
                            DynamicPartitionsSnap,
                        )
                        else None
                    ),
                )
                for dim in partition_def_data.partition_dimensions
            ]
            if isinstance(partition_def_data, MultiPartitionsSnap)
            else [
                GrapheneDimensionDefinitionType(
                    name="default",
                    description="",
                    type=GraphenePartitionDefinitionType.from_partition_def_data(
                        partition_def_data
                    ),
                    isPrimaryDimension=True,
                    dynamicPartitionsDefinitionName=(
                        partition_def_data.name
                        if isinstance(partition_def_data, DynamicPartitionsSnap)
                        else None
                    ),
                )
            ]
        )

    def __init__(self, partition_def_data: PartitionsSnap):
        self._partition_def_data = partition_def_data
        super().__init__(
            description=str(partition_def_data.get_partitions_definition()),
            type=GraphenePartitionDefinitionType.from_partition_def_data(partition_def_data),
            name=(
                partition_def_data.name
                if isinstance(partition_def_data, DynamicPartitionsSnap)
                else None
            ),
            fmt=(
                partition_def_data.fmt
                if isinstance(partition_def_data, TimeWindowPartitionsSnap)
                else None
            ),
        )


class GrapheneDimensionPartitionKeys(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    partition_keys = non_null_list(graphene.String)
    type = graphene.NonNull(GraphenePartitionDefinitionType)

    class Meta:
        name = "DimensionPartitionKeys"


types = [
    GraphenePartition,
    GraphenePartitionRunConfig,
    GraphenePartitionRunConfigOrError,
    GraphenePartitions,
    GraphenePartitionSet,
    GraphenePartitionSetOrError,
    GraphenePartitionSets,
    GraphenePartitionSetsOrError,
    GraphenePartitionsOrError,
    GraphenePartitionStatus,
    GraphenePartitionStatusCounts,
    GraphenePartitionStatuses,
    GraphenePartitionStatusesOrError,
    GraphenePartitionTags,
    GraphenePartitionTagsOrError,
]



from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional, Union

import dagster._check as check
from dagster._core.definitions.asset_key import AssetCheckKey, AssetKey, EntityKey
from dagster._core.scheduler.instigation import AutoMaterializeAssetEvaluationRecord

from dagster_graphql.schema.asset_condition_evaluations import (
    GrapheneAssetConditionEvaluation,
    GrapheneAssetConditionEvaluationRecord,
    GrapheneAssetConditionEvaluationRecords,
    GrapheneAssetConditionEvaluationRecordsOrError,
)
from dagster_graphql.schema.auto_materialize_asset_evaluations import (
    GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError,
)
from dagster_graphql.schema.inputs import GrapheneAssetCheckHandleInput, GrapheneAssetKeyInput

if TYPE_CHECKING:
    from dagster_graphql.schema.util import ResolveInfo


def _get_migration_error(
    graphene_info: "ResolveInfo",
) -> Optional[GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError]:
    if graphene_info.context.instance.schedule_storage is None:
        return GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError(
            message="Instance does not have schedule storage configured, cannot fetch evaluations."
        )
    if not graphene_info.context.instance.schedule_storage.supports_auto_materialize_asset_evaluations:
        return GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError(
            message=(
                "Auto materialize evaluations are not getting logged. Run `dagster instance"
                " migrate` to enable."
            )
        )
    return None


def _get_graphene_records_from_evaluations(
    graphene_info: "ResolveInfo",
    evaluation_records: Sequence[AutoMaterializeAssetEvaluationRecord],
) -> GrapheneAssetConditionEvaluationRecords:
    return GrapheneAssetConditionEvaluationRecords(
        records=[
            GrapheneAssetConditionEvaluationRecord(evaluation) for evaluation in evaluation_records
        ]
    )


def fetch_asset_condition_evaluation_record_for_partition(
    graphene_info: "ResolveInfo",
    graphene_asset_key: GrapheneAssetKeyInput,
    evaluation_id: int,
    partition_key: str,
) -> GrapheneAssetConditionEvaluation:
    asset_key = AssetKey.from_graphql_input(graphene_asset_key)
    schedule_storage = check.not_none(graphene_info.context.instance.schedule_storage)
    record = next(
        iter(
            schedule_storage.get_auto_materialize_asset_evaluations(
                asset_key, cursor=evaluation_id + 1, limit=1
            )
        )
    )
    return GrapheneAssetConditionEvaluation(
        record.get_evaluation_with_run_ids().evaluation, partition_key
    )


def entity_key_from_graphql_input(
    graphene_input: Union[GrapheneAssetKeyInput, GrapheneAssetCheckHandleInput],
) -> EntityKey:
    if "path" in graphene_input:
        return AssetKey.from_graphql_input(graphene_input)
    else:
        return AssetCheckKey.from_graphql_input(graphene_input)


def fetch_true_partitions_for_evaluation_node(
    graphene_info: "ResolveInfo",
    graphene_entity_key: Union[GrapheneAssetKeyInput, GrapheneAssetCheckHandleInput],
    evaluation_id: int,
    node_unique_id: str,
) -> Sequence[str]:
    key = entity_key_from_graphql_input(graphene_entity_key)
    schedule_storage = check.not_none(graphene_info.context.instance.schedule_storage)
    record = next(
        iter(
            schedule_storage.get_auto_materialize_asset_evaluations(
                # there is no method to get a specific evaluation by id, so instead get the first
                # evaluation before evaluation_id + 1
                key=key,
                cursor=evaluation_id + 1,
                limit=1,
            )
        )
    )
    check.invariant(
        record.evaluation_id == evaluation_id, f"Could not find evaluation with id {evaluation_id}"
    )

    # it's no longer necessary to pass in the partitions def in to get_evaluation_with_run_ids
    root_evaluation = record.get_evaluation_with_run_ids().evaluation
    for evaluation in root_evaluation.iter_nodes():
        if evaluation.condition_snapshot.unique_id == node_unique_id:
            return (
                list(evaluation.true_subset.subset_value.get_partition_keys())
                if evaluation.true_subset.is_partitioned
                else ["None"]
            )
    check.failed("No matching unique id found")


def fetch_asset_condition_evaluation_records_for_asset_key(
    graphene_info: "ResolveInfo",
    graphene_entity_key: Union[GrapheneAssetKeyInput, GrapheneAssetCheckHandleInput],
    limit: int,
    cursor: Optional[str],
) -> GrapheneAssetConditionEvaluationRecordsOrError:
    """Fetch asset policy evaluations from storage."""
    migration_error = _get_migration_error(graphene_info)
    if migration_error:
        return migration_error

    key = entity_key_from_graphql_input(graphene_entity_key)

    schedule_storage = check.not_none(graphene_info.context.instance.schedule_storage)
    return _get_graphene_records_from_evaluations(
        graphene_info,
        schedule_storage.get_auto_materialize_asset_evaluations(
            key=key,
            limit=limit,
            cursor=int(cursor) if cursor else None,
        ),
    )


def fetch_asset_condition_evaluation_records_for_evaluation_id(
    graphene_info: "ResolveInfo",
    evaluation_id: int,
) -> GrapheneAssetConditionEvaluationRecordsOrError:
    migration_error = _get_migration_error(graphene_info)
    if migration_error:
        return migration_error

    schedule_storage = check.not_none(graphene_info.context.instance.schedule_storage)
    return _get_graphene_records_from_evaluations(
        graphene_info,
        schedule_storage.get_auto_materialize_evaluations_for_evaluation_id(
            evaluation_id=evaluation_id
        ),
    )



import asyncio

import graphene

from dagster_graphql.implementation.fetch_asset_health import (
    get_asset_check_status_and_metadata,
    get_freshness_status_and_metadata,
    get_materialization_status_and_metadata,
)
from dagster_graphql.schema.util import ResolveInfo


class GrapheneAssetHealthStatus(graphene.Enum):
    HEALTHY = "HEALTHY"
    WARNING = "WARNING"
    DEGRADED = "DEGRADED"
    UNKNOWN = "UNKNOWN"
    NOT_APPLICABLE = "NOT_APPLICABLE"

    class Meta:
        name = "AssetHealthStatus"


class GrapheneAssetHealthCheckDegradedMeta(graphene.ObjectType):
    numFailedChecks = graphene.NonNull(graphene.Int)
    numWarningChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckDegradedMeta"


class GrapheneAssetHealthCheckWarningMeta(graphene.ObjectType):
    numWarningChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckWarningMeta"


class GrapheneAssetHealthCheckUnknownMeta(graphene.ObjectType):
    numNotExecutedChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckUnknownMeta"


class GrapheneAssetHealthCheckMeta(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetHealthCheckDegradedMeta,
            GrapheneAssetHealthCheckWarningMeta,
            GrapheneAssetHealthCheckUnknownMeta,
        )
        name = "AssetHealthCheckMeta"


class GrapheneAssetHealthMaterializationDegradedPartitionedMeta(graphene.ObjectType):
    numFailedPartitions = graphene.NonNull(graphene.Int)
    numMissingPartitions = graphene.NonNull(graphene.Int)
    totalNumPartitions = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthMaterializationDegradedPartitionedMeta"


class GrapheneAssetHealthMaterializationHealthyPartitionedMeta(graphene.ObjectType):
    numMissingPartitions = graphene.NonNull(graphene.Int)
    totalNumPartitions = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthMaterializationHealthyPartitionedMeta"


class GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta(graphene.ObjectType):
    failedRunId = graphene.NonNull(graphene.String)

    class Meta:
        name = "AssetHealthMaterializationDegradedNotPartitionedMeta"


class GrapheneAssetHealthMaterializationMeta(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetHealthMaterializationDegradedPartitionedMeta,
            GrapheneAssetHealthMaterializationHealthyPartitionedMeta,
            GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta,
        )
        name = "AssetHealthMaterializationMeta"


class GrapheneAssetHealthFreshnessMeta(graphene.ObjectType):
    lastMaterializedTimestamp = graphene.Field(graphene.Float)

    class Meta:
        name = "AssetHealthFreshnessMeta"


class GrapheneAssetHealth(graphene.ObjectType):
    assetHealth = graphene.NonNull(GrapheneAssetHealthStatus)
    materializationStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    materializationStatusMetadata = graphene.Field(GrapheneAssetHealthMaterializationMeta)
    assetChecksStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    assetChecksStatusMetadata = graphene.Field(GrapheneAssetHealthCheckMeta)
    freshnessStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    freshnessStatusMetadata = graphene.Field(GrapheneAssetHealthFreshnessMeta)

    class Meta:
        name = "AssetHealth"

    def __init__(self, asset_node_snap, dynamic_partitions_loader):
        super().__init__()
        self._asset_node_snap = asset_node_snap
        self._dynamic_partitions_loader = dynamic_partitions_loader
        self.materialization_status_task = None
        self.asset_check_status_task = None
        self.freshness_status_task = None

    async def resolve_materializationStatus(self, graphene_info: ResolveInfo) -> str:
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                get_materialization_status_and_metadata(
                    graphene_info, self._asset_node_snap.asset_key
                )
            )
        materialization_status, _ = await self.materialization_status_task
        return materialization_status

    async def resolve_materializationStatusMetadata(
        self, graphene_info: ResolveInfo
    ) -> GrapheneAssetHealthMaterializationMeta:
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                get_materialization_status_and_metadata(
                    graphene_info, self._asset_node_snap.asset_key
                )
            )
        _, materialization_status_metadata = await self.materialization_status_task
        return materialization_status_metadata

    async def resolve_assetChecksStatus(self, graphene_info: ResolveInfo) -> str:
        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                get_asset_check_status_and_metadata(graphene_info, self._asset_node_snap.asset_key)
            )

        asset_checks_status, _ = await self.asset_check_status_task
        return asset_checks_status

    async def resolve_assetChecksStatusMetadata(
        self, graphene_info: ResolveInfo
    ) -> GrapheneAssetHealthCheckMeta:
        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                get_asset_check_status_and_metadata(graphene_info, self._asset_node_snap.asset_key)
            )

        _, asset_checks_status_metadata = await self.asset_check_status_task
        return asset_checks_status_metadata

    async def resolve_freshnessStatus(self, graphene_info: ResolveInfo) -> str:
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                get_freshness_status_and_metadata(graphene_info, self._asset_node_snap.asset_key)
            )

        freshness_status, _ = await self.freshness_status_task
        return freshness_status

    async def resolve_freshnessStatusMetadata(
        self, graphene_info: ResolveInfo
    ) -> GrapheneAssetHealthFreshnessMeta:
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                get_freshness_status_and_metadata(graphene_info, self._asset_node_snap.asset_key)
            )

        _, freshness_status_metadata = await self.freshness_status_task
        return freshness_status_metadata

    async def resolve_assetHealth(self, graphene_info: ResolveInfo):
        if not graphene_info.context.instance.dagster_observe_supported():
            return GrapheneAssetHealthStatus.UNKNOWN
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                get_materialization_status_and_metadata(
                    graphene_info, self._asset_node_snap.asset_key
                )
            )
        materialization_status, _ = await self.materialization_status_task
        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                get_asset_check_status_and_metadata(graphene_info, self._asset_node_snap.asset_key)
            )
        asset_checks_status, _ = await self.asset_check_status_task
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                get_freshness_status_and_metadata(graphene_info, self._asset_node_snap.asset_key)
            )
        freshness_status, _ = await self.freshness_status_task
        statuses = [
            materialization_status,
            asset_checks_status,
            freshness_status,
        ]
        if GrapheneAssetHealthStatus.DEGRADED in statuses:
            return GrapheneAssetHealthStatus.DEGRADED
        if GrapheneAssetHealthStatus.WARNING in statuses:
            return GrapheneAssetHealthStatus.WARNING
        # at this point, all statuses are HEALTHY, UNKNOWN, or NOT_APPLICABLE
        if materialization_status == GrapheneAssetHealthStatus.UNKNOWN:
            return GrapheneAssetHealthStatus.UNKNOWN
        if all(
            status == GrapheneAssetHealthStatus.UNKNOWN
            or status == GrapheneAssetHealthStatus.NOT_APPLICABLE
            for status in statuses
        ):
            return GrapheneAssetHealthStatus.UNKNOWN
        # at least one status must be HEALTHY
        return GrapheneAssetHealthStatus.HEALTHY



import graphene

from dagster_graphql.schema.roots.mutation import GrapheneMutation
from dagster_graphql.schema.roots.query import GrapheneQuery
from dagster_graphql.schema.roots.subscription import GrapheneSubscription


def types():
    from dagster_graphql.schema.backfill import (
        GrapheneLaunchBackfillResult,
        GrapheneLaunchBackfillSuccess,
    )
    from dagster_graphql.schema.config_type_or_error import GrapheneConfigTypeOrError
    from dagster_graphql.schema.config_types import types as config_types
    from dagster_graphql.schema.dagster_types import types as dagster_types_types
    from dagster_graphql.schema.entity_key import GrapheneAssetKey
    from dagster_graphql.schema.errors import types as errors_types
    from dagster_graphql.schema.execution import types as execution_types
    from dagster_graphql.schema.external import types as external_types
    from dagster_graphql.schema.inputs import types as inputs_types
    from dagster_graphql.schema.instance import (
        GrapheneDaemonHealth,
        GrapheneDaemonStatus,
        GrapheneInstance,
        GrapheneRunLauncher,
    )
    from dagster_graphql.schema.instigation import types as instigation_types
    from dagster_graphql.schema.logs import types as log_types
    from dagster_graphql.schema.metadata import types as metadata_types
    from dagster_graphql.schema.partition_sets import types as partition_sets_types
    from dagster_graphql.schema.pipelines import types as pipelines_types
    from dagster_graphql.schema.repository_origin import (
        GrapheneRepositoryMetadata,
        GrapheneRepositoryOrigin,
    )
    from dagster_graphql.schema.roots import types as roots_types
    from dagster_graphql.schema.run_config import (
        GrapheneRunConfigSchema,
        GrapheneRunConfigSchemaOrError,
    )
    from dagster_graphql.schema.runs import types as runs_types
    from dagster_graphql.schema.schedules import types as schedules_types
    from dagster_graphql.schema.sensors import types as sensors_types
    from dagster_graphql.schema.solids import types as solids_types
    from dagster_graphql.schema.table import types as table_types
    from dagster_graphql.schema.tags import GraphenePipelineTag, GraphenePipelineTagAndValues
    from dagster_graphql.schema.used_solid import GrapheneNodeInvocationSite, GrapheneUsedSolid

    return (
        log_types()
        + pipelines_types()
        + roots_types()
        + schedules_types()
        + [GrapheneAssetKey]
        + [GrapheneLaunchBackfillResult, GrapheneLaunchBackfillSuccess]
        + [GrapheneConfigTypeOrError]
        + config_types
        + dagster_types_types
        + errors_types
        + execution_types
        + external_types
        + inputs_types
        + [GrapheneDaemonHealth, GrapheneDaemonStatus, GrapheneInstance, GrapheneRunLauncher]
        + instigation_types
        + metadata_types()
        + partition_sets_types
        + [GrapheneRepositoryOrigin, GrapheneRepositoryMetadata]
        + [GrapheneRunConfigSchema, GrapheneRunConfigSchemaOrError]
        + runs_types
        + sensors_types
        + solids_types
        + table_types
        + [GraphenePipelineTag, GraphenePipelineTagAndValues]
        + [GrapheneNodeInvocationSite, GrapheneUsedSolid]
    )


def create_schema() -> graphene.Schema:
    return graphene.Schema(
        query=GrapheneQuery,
        mutation=GrapheneMutation,
        subscription=GrapheneSubscription,
        types=types(),
    )



from typing import TYPE_CHECKING, Optional

from dagster._core.execution.backfill import BulkActionsFilter, BulkActionStatus

if TYPE_CHECKING:
    from dagster_graphql.schema.backfill import (
        GraphenePartitionBackfill,
        GraphenePartitionBackfills,
    )
    from dagster_graphql.schema.util import ResolveInfo


def get_backfill(graphene_info: "ResolveInfo", backfill_id: str) -> "GraphenePartitionBackfill":
    from dagster_graphql.schema.backfill import (
        GrapheneBackfillNotFoundError,
        GraphenePartitionBackfill,
    )

    backfill_job = graphene_info.context.instance.get_backfill(backfill_id)
    if backfill_job is None:
        return GrapheneBackfillNotFoundError(backfill_id)

    return GraphenePartitionBackfill(backfill_job)


def get_backfills(
    graphene_info: "ResolveInfo",
    filters: Optional[BulkActionsFilter] = None,
    cursor: Optional[str] = None,
    limit: Optional[int] = None,
    status: Optional[BulkActionStatus] = None,
) -> "GraphenePartitionBackfills":
    from dagster_graphql.schema.backfill import (
        GraphenePartitionBackfill,
        GraphenePartitionBackfills,
    )

    backfills = graphene_info.context.instance.get_backfills(
        status=status, cursor=cursor, limit=limit, filters=filters
    )
    return GraphenePartitionBackfills(
        results=[GraphenePartitionBackfill(backfill) for backfill in backfills]
    )



from dagster_graphql.client.client import DagsterGraphQLClient as DagsterGraphQLClient
from dagster_graphql.client.utils import (
    DagsterGraphQLClientError as DagsterGraphQLClientError,
    InvalidOutputErrorInfo as InvalidOutputErrorInfo,
    ReloadRepositoryLocationInfo as ReloadRepositoryLocationInfo,
    ReloadRepositoryLocationStatus as ReloadRepositoryLocationStatus,
    ShutdownRepositoryLocationInfo as ShutdownRepositoryLocationInfo,
    ShutdownRepositoryLocationStatus as ShutdownRepositoryLocationStatus,
)



from typing import Optional

from dagster import _check as check
from dagster._core.definitions.partition import CachingDynamicPartitionsLoader, PartitionsSubset
from dagster._core.remote_representation.external_data import AssetNodeSnap
from dagster._core.storage.partition_status_cache import get_partition_subsets
from dagster._core.workspace.context import WorkspaceRequestContext


def regenerate_and_check_partition_subsets(
    context: WorkspaceRequestContext,
    asset_node_snap: AssetNodeSnap,
    dynamic_partitions_loader: Optional[CachingDynamicPartitionsLoader],
) -> tuple[PartitionsSubset[str], PartitionsSubset[str], PartitionsSubset[str]]:
    if not dynamic_partitions_loader:
        check.failed("dynamic_partitions_loader must be provided to get partition keys")

    (
        materialized_partition_subset,
        failed_partition_subset,
        in_progress_subset,
    ) = get_partition_subsets(
        context.instance,
        context,
        asset_node_snap.asset_key,
        dynamic_partitions_loader,
        (
            asset_node_snap.partitions.get_partitions_definition()
            if asset_node_snap.partitions
            else None
        ),
    )

    if (
        materialized_partition_subset is None
        or failed_partition_subset is None
        or in_progress_subset is None
    ):
        check.failed("Expected partitions subset for a partitioned asset")

    return materialized_partition_subset, failed_partition_subset, in_progress_subset
